#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/interrupt.h>

#include <asm/mach/pci.h>
#include <asm/arch/ns9750_pci.h>

#define PCI_NAME "ns9750_pci"

#undef DEBUG
//#define DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC() printk(PCI_NAME ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...) printk(PCI_NAME ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif

#define PRINT_WARN(args...) printk(KERN_WARNING PCI_NAME ": " args)

#define NS9750_PCI_IO_BASE  (NS_PCI_IO_BASE - IO_BASE)
#define NS9750_PCI_IO_SIZE  NS_PCI_IO_SIZE
#define NS9750_PCI_MEM_BASE NS_PCI_MEM_BASE_PA
#define NS9750_PCI_MEM_SIZE NS_PCI_MEM_SIZE

#define CONFIG_CMD(bus, devfn, where) \
	(0x80000000 | (bus->number << 16) | (devfn << 8) | (where & ~3))

static int
ns9750_read_config(struct pci_bus *bus, unsigned int devfn, int where,
                      int size, u32 *value)
{
	if(devfn & 0xf0){
		*value = 0;
		return PCIBIOS_SUCCESSFUL;
	}
	writel(CONFIG_CMD(bus, devfn, where), NS_PCI_CONFIG_ADDR_BASE);
	switch(size){
	case 1:
		*value = readb(NS_PCI_CONFIG_DATA_BASE + (where & 3));
		break;
	case 2:
		*value = readw(NS_PCI_CONFIG_DATA_BASE + (where & 2));
		break;
	case 4:
		*value = readl(NS_PCI_CONFIG_DATA_BASE);
		break;
	default:
		*value = 0;
		break;
	}
	DEBUG_INFO("[RD:%d] *0x%08x = 0x%08x\n",
		   size, CONFIG_CMD(bus, devfn, where) + (where & 3), *value);
	return PCIBIOS_SUCCESSFUL;
}

static int
ns9750_write_config(struct pci_bus *bus, unsigned int devfn, int where,
                       int size, u32 value)
{
	if(devfn & 0xf0){
		return PCIBIOS_SUCCESSFUL;
	}

	writel(CONFIG_CMD(bus, devfn, where), NS_PCI_CONFIG_ADDR_BASE);
	switch(size){
	case 1:
		writeb(value, NS_PCI_CONFIG_DATA_BASE + (where & 3));
		break;
	case 2:
		writew(value, NS_PCI_CONFIG_DATA_BASE + (where & 2));
		break;
	case 4:
		writel(value, NS_PCI_CONFIG_DATA_BASE);
		break;
	default:
		break;
	}
	DEBUG_INFO("[WR:%d] *0x%08x = 0x%08x\n",
		   size, CONFIG_CMD(bus, devfn, where) + (where & 3), value);
	return PCIBIOS_SUCCESSFUL;
}

#define SLOT_NUM 2
#define PIN_NUM  3
static int irq_table[SLOT_NUM][PIN_NUM] __initdata = {
	{ -1, -1, -1 },
	{ -1, IRQ_PCI_EXT_0, IRQ_PCI_EXT_1 },
};

static int __init
ns9750_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	DEBUG_FUNC();
	if (SLOT_NUM <= slot || PIN_NUM <= pin)
		return -1;
	return irq_table[slot][pin];
}

static struct pci_ops ns9750_ops = {
	.read	= ns9750_read_config,
	.write	= ns9750_write_config,
};

static struct pci_bus * __init
ns9750_scan_bus(int nr, struct pci_sys_data *sys)
{
	DEBUG_FUNC();
	return pci_scan_bus(0, &ns9750_ops, sys);
}

static struct resource io_resource = {
	.name	= "PCI io",
	.start	= NS9750_PCI_IO_BASE,
	.end	= NS9750_PCI_IO_BASE + NS9750_PCI_IO_SIZE - 1,
	.flags	= IORESOURCE_IO,
};

static struct resource mem_resource = {
	.name	= "PCI mem",
	.start	= NS9750_PCI_MEM_BASE,
	.end	= NS9750_PCI_MEM_BASE + NS9750_PCI_MEM_SIZE - 1,
	.flags	= IORESOURCE_MEM,
};

static int __init
ns9750_setup(int nr, struct pci_sys_data *sys)
{
	DEBUG_FUNC();

	if (!request_mem_region(NS_PCI_CONFIG_ADDR_BASE_PA,
				NS_PCI_CONFIG_ADDR_SIZE,
				"pci config address")){
		PRINT_WARN("Unable to request region pci config address\n");
		return 0;
	}
	if (!request_mem_region(NS_PCI_CONFIG_DATA_BASE_PA,
				NS_PCI_CONFIG_DATA_SIZE, "pci config data")){
		PRINT_WARN("Unable to request region pci config data\n");
		release_region(NS_PCI_CONFIG_ADDR_BASE_PA,
			       NS_PCI_CONFIG_ADDR_SIZE);
		return 0;
	}

	sys->resource[0] = &io_resource;
	sys->resource[1] = &mem_resource;
	sys->resource[2] = NULL;

	return 1;
}

static void __init
ns9750_preinit(void)
{
	DEBUG_FUNC();

	writel( NS_PCI_EN_BAR0, NS_PCI_PMISC );

	/* Set up bridge configuration in CMD register */
	writel( 0x80000004, NS_PCI_CONFIG_ADDR_BASE );
	writel( 0x00000046, NS_PCI_CONFIG_DATA_BASE );

	/* config BAR0: 0x00000000, 256 Mbytes */
	writel( 0x80000010, NS_PCI_CONFIG_ADDR_BASE );
	writel( 0x00000000, NS_PCI_CONFIG_DATA_BASE );

	/* Set up memory space translation */
	writel( (((NS9750_PCI_MEM_BASE >> 25) | 0x3) << 24) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x2) << 16) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x1) <<  8) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x0) <<  0),
		NS_PCI_PALTMEM0 );
	writel( (((NS9750_PCI_MEM_BASE >> 25) | 0x7) << 24) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x6) << 16) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x5) <<  8) |
		(((NS9750_PCI_MEM_BASE >> 25) | 0x4) <<  0),
		NS_PCI_PALTMEM1 );

	writel( NS9750_PCI_IO_BASE >> 20, NS_PCI_PALTIO );
	writel( 0x00000000 >> 28, NS_PCI_PMALT0 );
	writel( 0x3, NS_PCI_PALTCTL );

	/* config latency and cache size registers */
	writel( 0x8000000C, NS_PCI_CONFIG_ADDR_BASE );
	writel( 0x0000FF00, NS_PCI_CONFIG_DATA_BASE );
}

static irqreturn_t
ns9750_pci_arbiter_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 status;

	status = readl(NS_PCI_PARBINT) & 0xe;
	if (status) {
		writel(status, NS_PCI_PARBINT);
		writel(readl(NS_PCI_PARBCFG) | status, NS_PCI_PARBCFG);
		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}

static void __init
ns9750_postinit(void)
{
	DEBUG_FUNC();

	request_irq(IRQ_PCI_ARBITER, ns9750_pci_arbiter_interrupt,
		    SA_INTERRUPT, "PCI Arbiter", NULL);
	writel( 0x2, NS_PCI_PARBINTEN );
	writel( 0x2, NS_PCI_PARBCFG );
}

static struct hw_pci ns9750_pci __initdata = {
	.nr_controllers	= 1,
	.swizzle	= pci_std_swizzle,
	.setup		= ns9750_setup,
	.map_irq	= ns9750_map_irq,
	.scan		= ns9750_scan_bus,
	.preinit	= ns9750_preinit,
	.postinit	= ns9750_postinit,
};

static int __init ns9750_pci_init(void)
{
	DEBUG_FUNC();

	pci_common_init(&ns9750_pci);

	return 0;
}

subsys_initcall(ns9750_pci_init);
