/*
 * Handle mapping of the NOR flash on Armadillo-3x0 board
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/init.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#include <asm/mach/flash.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>

struct armadillo3x0_flash_info {
        struct flash_platform_data *plat;
        struct resource         *res;
        struct mtd_partition    *parts;
        struct mtd_info         *mtd;
        struct map_info         map;
};

#define MAP_NAME "armadillo3x0-nor"

#define PRINT_ERR(args...)  printk( KERN_ERR  MAP_NAME ": " args)
#define PRINT_INFO(args...) printk( KERN_INFO MAP_NAME ": " args)

static struct mtd_partition armadillo3x0_partitions[] = {
	{
		.name		= "ipl",
		.offset		= 0x00000000,
		.size		= 0x00002000,
		.mask_flags	= MTD_WRITEABLE, /* force read-only */
	},
	{
		.name		= "bootloader",
		.offset		= 0x00002000,
		.size		= 0x0000e000,
		.mask_flags	= MTD_WRITEABLE, /* force read-only */
	},
	{
		.name		= "kernel",
		.offset		= 0x00010000,
		.size		= 0x00200000,
	},
	{
		.name		= "userland",
		.offset		= 0x00210000,
		.size		= 0x005e0000,
	},
	{
		.name		= "config",
		.offset		= 0x007f0000,
		.size		= 0x00010000,
	},
};

static const char *probes[] = { "cmdlinepart", NULL };

static int
armadillo3x0_flash_probe(struct device *_dev)
{
  	struct platform_device *dev = to_platform_device(_dev);
	struct flash_platform_data *plat = dev->dev.platform_data;
	struct resource *res = dev->resource;
	unsigned int size = res->end - res->start + 1;
	struct armadillo3x0_flash_info *info;
	int err;
	void *base;
	int default_parts = 0;

	info = kmalloc(sizeof(struct armadillo3x0_flash_info), GFP_KERNEL);
	if (!info) {
		err = -ENOMEM;
		goto out;
	}

	memset(info, 0, sizeof(struct armadillo3x0_flash_info));

	info->plat = plat;
	if (plat && plat->init) {
		err = plat->init();
		if (err)
			goto no_resource;
	}

	info->res = request_mem_region(res->start, size, "mtd");
	if (!info->res) {
		err = -EBUSY;
		goto no_resource;
	}

	base = ioremap(res->start, size);
	if (!base) {
		err = -ENOMEM;
		goto no_mem;
	}

	/*
	 * look for CFI based flash parts fitted to this board
	 */
	info->map.size		= size;
	info->map.bankwidth	= plat->width;
	info->map.phys		= res->start;
	info->map.virt		= base;
	info->map.name		= dev->name;
	info->map.set_vpp	= 0;

	simple_map_init(&info->map);

	/*
	 * Also, the CFI layer automatically works out what size
	 * of chips we have, and does the necessary identification
	 * for us automatically.
	 */
	info->mtd = do_map_probe(plat->map_name, &info->map);
	if (!info->mtd) {
		err = -ENXIO;
		goto no_device;
	}

	info->mtd->owner = THIS_MODULE;

	err = parse_mtd_partitions(info->mtd, probes, &info->parts, 0);
	if (err > 0) {
		PRINT_INFO("use cmdline partitions(%d)\n", err);
		err = add_mtd_partitions(info->mtd, info->parts, err);
		if (err) {
			PRINT_ERR("mtd partition registration failed: %d\n",
				  err);
			default_parts = 1;
		}
	} else {
		default_parts = 1;
	}

	if (default_parts) {
		int partno = \
			sizeof(armadillo3x0_partitions) / \
			sizeof(armadillo3x0_partitions[1]);
		PRINT_INFO("use default partitions(%d)\n", partno);
		err = add_mtd_partitions(info->mtd, 
					 armadillo3x0_partitions, partno);
		if (err)
			PRINT_ERR("mtd partition registration failed: %d\n",
				  err);
	}
	
	if (err == 0)
		dev_set_drvdata(&dev->dev, info);

	/*
	 * If we got an error, free all resources.
	 */
	if (err < 0) {
		if (info->mtd) {
			del_mtd_partitions(info->mtd);
			map_destroy(info->mtd);
		}
		if (info->parts)
			kfree(info->parts);

 no_device:
		iounmap(base);
 no_mem:
		release_mem_region(res->start, size);
 no_resource:
		if (plat && plat->exit)
			plat->exit();
		kfree(info);
	}
 out:
	return err;
}

static int
armadillo3x0_flash_remove(struct device *_dev)
{
	struct platform_device *dev = to_platform_device(_dev);
	struct armadillo3x0_flash_info *info = dev_get_drvdata(&dev->dev);

	dev_set_drvdata(&dev->dev, NULL);

	if (info) {
		if (info->mtd) {
			del_mtd_partitions(info->mtd);
			map_destroy(info->mtd);
		}
		if (info->parts)
			kfree(info->parts);

		iounmap((void *)info->map.virt);
		release_resource(info->res);
		kfree(info->res);

		if (info->plat && info->plat->exit)
			info->plat->exit();

		kfree(info);
	}

	return 0;
}


static struct device_driver armadillo3x0_flash_driver = {
	.name	= MAP_NAME,
	.bus	= &platform_bus_type,
	.probe	= armadillo3x0_flash_probe,
	.remove	= armadillo3x0_flash_remove,
};

static int __init 
armadillo3x0_flash_init(void)
{
	return driver_register(&armadillo3x0_flash_driver);
}

static void __exit 
armadillo3x0_flash_exit(void)
{
	driver_unregister(&armadillo3x0_flash_driver);
}

module_init(armadillo3x0_flash_init);
module_exit(armadillo3x0_flash_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("Armadillo-3x0 CFI map driver");
